#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "lib.h"
#include "command.h"
#include "value.h"
#include "variable.h"
#include "label.h"
#include "array.h"
#include "function.h"
#include "position.h"
#include "nll.h"
#include "formula.h"

struct operator operators[] = {
  { "++" , 1, OPERATOR_TYPE_INC      , OPERATOR_SUBTYPE_MONADIC, OPERATOR_DIRECTION_RL   },
  { "--" , 1, OPERATOR_TYPE_DEC      , OPERATOR_SUBTYPE_MONADIC, OPERATOR_DIRECTION_RL   },
  { "<<",  4, OPERATOR_TYPE_LSHIFT   , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { ">>",  4, OPERATOR_TYPE_RSHIFT   , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "<=",  5, OPERATOR_TYPE_LE       , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { ">=",  5, OPERATOR_TYPE_GE       , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "==",  6, OPERATOR_TYPE_EQ       , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "!=",  6, OPERATOR_TYPE_NE       , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "<>",  6, OPERATOR_TYPE_NE       , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "&&", 10, OPERATOR_TYPE_LOGAND   , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "||", 11, OPERATOR_TYPE_LOGOR    , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "<-", 12, OPERATOR_TYPE_REFER    , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "!" ,  1, OPERATOR_TYPE_NOT      , OPERATOR_SUBTYPE_MONADIC, OPERATOR_DIRECTION_RL   },
  { "~" ,  1, OPERATOR_TYPE_INV      , OPERATOR_SUBTYPE_MONADIC, OPERATOR_DIRECTION_RL   },
  { "+" ,  1, OPERATOR_TYPE_PLUS     , OPERATOR_SUBTYPE_MON_BIN, OPERATOR_DIRECTION_RL   },
  { "-" ,  1, OPERATOR_TYPE_MINUS    , OPERATOR_SUBTYPE_MON_BIN, OPERATOR_DIRECTION_RL   },
  { "*" ,  1, OPERATOR_TYPE_POINTER  , OPERATOR_SUBTYPE_MON_BIN, OPERATOR_DIRECTION_RL   },
  { "&" ,  1, OPERATOR_TYPE_ADDRESS  , OPERATOR_SUBTYPE_MON_BIN, OPERATOR_DIRECTION_RL   },
  { "*" ,  2, OPERATOR_TYPE_MUL      , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "/" ,  2, OPERATOR_TYPE_DIV      , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "%" ,  2, OPERATOR_TYPE_MOD      , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "+" ,  3, OPERATOR_TYPE_ADD      , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "-" ,  3, OPERATOR_TYPE_SUB      , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "<" ,  5, OPERATOR_TYPE_LT       , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { ">" ,  5, OPERATOR_TYPE_GT       , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "^" ,  8, OPERATOR_TYPE_XOR      , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "&" ,  7, OPERATOR_TYPE_AND      , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "|" ,  9, OPERATOR_TYPE_OR       , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "=" , 12, OPERATOR_TYPE_SUBST    , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_RL   },
  { "[" ,  0, OPERATOR_TYPE_ARRAY    , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { "," , 13, OPERATOR_TYPE_COMMA    , OPERATOR_SUBTYPE_NONE   , OPERATOR_DIRECTION_NONE },
  { NULL, 13, OPERATOR_TYPE_CONT     , OPERATOR_SUBTYPE_BINARY , OPERATOR_DIRECTION_LR   },
  { NULL, 13, OPERATOR_TYPE_NOP      , OPERATOR_SUBTYPE_NONE   , OPERATOR_DIRECTION_NONE },
  { NULL,  1, OPERATOR_TYPE_FUNCTION , OPERATOR_SUBTYPE_MONADIC, OPERATOR_DIRECTION_RL   },
  { NULL,  1, OPERATOR_TYPE_LABELCALL, OPERATOR_SUBTYPE_MONADIC, OPERATOR_DIRECTION_RL   },
  { NULL,  0, OPERATOR_TYPE_NONE     , OPERATOR_SUBTYPE_NONE   , OPERATOR_DIRECTION_NONE },
};

#define OPERATOR_PRIORITY_MIN  0
#define OPERATOR_PRIORITY_MAX 13
#define OPERATOR_PRIORITY_NUM \
	(OPERATOR_PRIORITY_MAX - OPERATOR_PRIORITY_MIN + 1)

static struct element elements[ELEMENT_MAXNUM];
static element_t pool = NULL;

int formula_check(void)
{
  element_t element;
  int n = 0;

  for (element = pool; element && (n < ELEMENT_MAXNUM); element = element->next)
    n++;

  if (element || (n != ELEMENT_MAXNUM))
    return NLL_ERRCODE_FORMULA_NOT_EMPTY;

  return 0;
}

static operator_t operator_get(operator_type_t type)
{
  operator_t operator;
  for (operator = operators; operator->type != OPERATOR_TYPE_NONE; operator++) {
    if (operator->type == type)
      return operator;
  }
  return NULL;
}

static int element_alloc(element_t *elementp)
{
  element_t element;

  if (!pool)
    return NLL_ERRCODE_FORMULA_BUFFER_OVER;

  element = pool;
  pool = pool->next;

  memset(element, 0, sizeof(*element));
  element->next = NULL;
  element->type = ELEMENT_TYPE_NONE;

  *elementp = element;

  return 0;
}

static int element_free(element_t element)
{
  element_t next;
  int r;

  for (; element; element = next) {
    next = element->next;

    if (element->value) {
      if ((r = value_free(element->value)) < 0)
	return r;
      element->value = NULL;
    }
    element->type = ELEMENT_TYPE_NONE;

    element->next = pool;
    pool = element;
  }

  return 0;
}

static int element_clean(element_t element)
{
  value_t value;
  int r;

  for (value = element->value; value; value = value->next) {
    if ((r = value_clear(value)) < 0)
      return r;
  }

  return 0;
}

static int element_init(void)
{
  element_t element;
  int i;

  pool = NULL;
  memset(elements, 0, sizeof(elements));

  for (i = 0; i < ELEMENT_MAXNUM; i++) {
    element = &elements[i];
    element->next = pool;
    pool = element;
  }

  return 0;
}

int formula_init(void)
{
  return element_init();
}

union val {
  integer_t integer;
  struct {
    char *s;
    int length;
  } string;
#ifdef NLL_FLOATING_POINT
  double floating;
#endif
};

static int element_alloc_values(element_t element, int num)
{
  int r;
  value_t value;

  while (element->value_num < num) {
    if ((r = value_alloc(&value)) < 0)
      return r;
    value->next = element->value;
    element->value = value;
    element->value_num++;
  }

  return 0;
}

static int element_push(element_t *elementp, element_type_t type,
			union val *val, element_param_t param)
{
  element_t element = NULL;
  int r;

  if ((r = element_alloc(&element)) < 0)
    goto err;

  element->type = type;
  switch (type) {
  case ELEMENT_TYPE_INTEGER:
    if ((r = element_alloc_values(element, 1)) < 0)
      goto err;
    if ((r = value_set_integer(element->value, val->integer)) < 0)
      goto err;
    if ((r = value_set_const(element->value)) < 0)
      goto err;
    break;
  case ELEMENT_TYPE_STRING:
    if ((r = element_alloc_values(element, 1)) < 0)
      goto err;
    if ((r = value_set_string(element->value, val->string.s, val->string.length)) < 0)
      goto err;
    if ((r = value_set_const(element->value)) < 0)
      goto err;
    break;
#ifdef NLL_FLOATING_POINT
  case ELEMENT_TYPE_FLOAT:
    if ((r = element_alloc_values(element, 1)) < 0)
      goto err;
    if ((r = value_set_float(element->value, val->floating)) < 0)
      goto err;
    if ((r = value_set_const(element->value)) < 0)
      goto err;
    break;
#endif
  default:
    break;
  }

  if (param)
    memcpy(&element->param, param, sizeof(element->param));

  element->next = *elementp;
  *elementp = element;

  return 0;

err:
  if (element)
    element_free(element);
  return r;
}

static int a2val(char c)
{
  if (isdigit(c))
    return c - '0';
  else if (isalpha(c))
    return toupper(c) - 'A' + 10;
  return -1;
}

static int lex(element_t *elementp, char *formula, char **endp, const char *terminator)
{
  int r, i, v, scale, is_float;
  integer_t integer;
  union val val;
#ifdef NLL_FLOATING_POINT
  double fscale;
#endif
  char quat = '\0';
  int escape = 0, escape_scale = 0;
  element_t head, *tail, element;
  char *p;
  char word[STRING_MAXLEN + 1], *wp = NULL;
  union element_param param;
  operator_t operator;

  p = formula;

  head = NULL;
  tail = &head;

  while (1) {
    if (*p == '\0')
      break;

    if (quat) {
      if (escape && escape_scale) {
	v = a2val(*p);
	if ((v < 0) || (v >= escape_scale)) {
	  *(wp++) = integer;
	  escape = 0;
	  p--;
	} else {
	  integer = integer * escape_scale + v;
	}
      } else if (escape) {
	switch (*p) {
	case '0': integer = 0; escape_scale =  8; break;
	case 'x':
	case 'X': integer = 0; escape_scale = 16; break;
	case 'b':
	case 'B': integer = 0; escape_scale =  2; break;
	case 'T':
	case 't': *(wp++) = '\t'; escape = 0; break;
	case 'N':
	case 'n': *(wp++) = '\n'; escape = 0; break;
	case 'V':
	case 'v': *(wp++) = '\v'; escape = 0; break;
	case 'R':
	case 'r': *(wp++) = '\r'; escape = 0; break;
	default:  *(wp++) = *p;   escape = 0; break;
	}
      } else if (*p == quat) {
	*wp = '\0';
	switch (quat) {
	case '\'':
	  val.integer = word[0];
	  if ((r = element_push(tail, ELEMENT_TYPE_INTEGER, &val, NULL)) < 0)
	    goto ret;
	  tail = &((*tail)->next);
	  break;
	case '\"':
	  val.string.s = word;
	  val.string.length = wp - word;
	  if ((r = element_push(tail, ELEMENT_TYPE_STRING, &val, NULL)) < 0)
	    goto ret;
	  tail = &((*tail)->next);
	  break;
	default:
	  break;
	}
	quat = '\0';
      } else if (*p == '\\') {
	escape = 1;
	escape_scale = 0;
      } else {
	*(wp++) = *p;
      }
      p++;
    } else if (strchr(terminator, *p)) {
      break;
    } else if ((*p == '\'') || (*p == '\"')) {
      quat = *p;
      escape = 0;
      wp = word;
      p++;
    } else if (isspace(*p)) {
      p++;
      continue;
    } else if (isdigit(*p)) { /* integer or floating */
      integer = 0;
      scale = 10;
      is_float = 0;
      if (*p == '0') {
	scale = 8;
	p++;
	if ((*p == 'x') || (*p == 'X')) {
	  scale = 16;
	  p++;
	} else if ((*p == 'b') || (*p == 'B')) {
	  scale = 2;
	  p++;
	}
      }
      while (1) {
#ifdef NLL_FLOATING_POINT
	if (*p == '.') {
	  val.floating = integer;
	  scale = 10; /* 0.1 is octal by top '0' */
	  integer = 0;
	  fscale = 1;
	  is_float = 1;
	  p++;
	  continue;
	}
#endif
	v = a2val(*p);
	if (v < 0)
	  break;
	if (v >= scale)
	  break;
	integer = integer * scale + v;
#ifdef NLL_FLOATING_POINT
	if (is_float)
	  fscale *= scale;
#endif
	p++;
      }
      if (is_float) {
#ifdef NLL_FLOATING_POINT
	val.floating += integer / fscale;
	if ((r = element_push(tail, ELEMENT_TYPE_FLOAT, &val, NULL)) < 0)
	  goto ret;
#endif
      } else {
	val.integer = integer;
	if ((r = element_push(tail, ELEMENT_TYPE_INTEGER, &val, NULL)) < 0)
	  goto ret;
      }
      tail = &((*tail)->next);
    } else if (isalpha(*p) || (*p == '_')) { /* variable or function name */
      for (i = 0; (isalnum(*p) || (*p == '_')); i++, p++) {
	*p = toupper(*p);
	param.name.name[i] = *p;
      }
      param.name.name[i] = '\0';
      if (*p == ':') {
	p++;
	if ((r = element_push(tail, ELEMENT_TYPE_LABEL, NULL, &param)) < 0)
	  goto ret;
      } else {
	if ((r = element_push(tail, ELEMENT_TYPE_NAME, NULL, &param)) < 0)
	  goto ret;
      }
      tail = &((*tail)->next);
    } else if (*p == '(') {
      p++;
      if ((r = lex(&element, p, &p, ")")) < 0)
	goto ret;
      if (*p) p++;
      param.elements.head = element;
      if ((r = element_push(tail, ELEMENT_TYPE_ELEMENTS, NULL, &param)) < 0)
	goto ret;
      tail = &((*tail)->next);
    } else { /* operator */
      for (operator = operators; operator->type != OPERATOR_TYPE_NONE; operator++) {
	if (operator->word && !strncmp(operator->word, p, strlen(operator->word)))
	  break;
      }
      if (operator->type == OPERATOR_TYPE_NONE) {
	r = NLL_ERRCODE_FORMULA_INVALID_OPERATOR;
	goto ret;
      }
      p += strlen(operator->word);
      param.operator.op = operator;
      param.operator.function = NULL;
      param.operator.label    = NULL;
      param.operator.args[0] = NULL;
      param.operator.args[1] = NULL;
      if ((r = element_push(tail, ELEMENT_TYPE_OPERATOR, NULL, &param)) < 0)
	goto ret;
      tail = &((*tail)->next);
      if (operator->type == OPERATOR_TYPE_ARRAY) {
	if ((r = lex(&element, p, &p, "]")) < 0)
	  goto ret;
	if (*p) p++;
	param.elements.head = element;
	if ((r = element_push(tail, ELEMENT_TYPE_ELEMENTS, NULL, &param)) < 0)
	  goto ret;
	tail = &((*tail)->next);
      }
    }
  }

  r = 0;

ret:
  *elementp = head;
  if (endp)
    *endp = p;

  return r;
}

static int elements_num(element_t element)
{
  int num;
  for (num = 0; element; element = element->next)
    num++;
  return num;
}

static int element_set_values(element_t element, element_t src)
{
  int r, i;
  value_t value;

  value = element->value;
  for (i = 0; i < element->value_num; i++) {
    if (!value || !src)
      return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
    if (src->value) {
      if ((r = value_set_value(value, src->value)) < 0)
	return r;
    }
    value = value->next;
    src = src->next;
  }

  return 0;
}

static int parse(element_t *elementp)
{
  int r, priority, i;
  element_t e, el, er, *elp;
  operator_direction_t d;
  operator_type_t opetype;
  function_t function;
  variable_t variable;
  label_t label;

  for (e = *elementp; e; e = e->next) {
    switch (e->type) {
    case ELEMENT_TYPE_ELEMENTS:
      if ((r = parse(&e->param.elements.head)) < 0)
	return r;
      if ((r = element_alloc_values(e, elements_num(e->param.elements.head))) < 0)
	return r;
      break;

    case ELEMENT_TYPE_LABEL:
      if ((r = label_set(e->param.name.name, NULL, &label)) < 0)
	return r;
      e->type = ELEMENT_TYPE_OPERATOR;
      e->param.operator.op = operator_get(OPERATOR_TYPE_LABELCALL);
      e->param.operator.function = NULL;
      e->param.operator.label    = label;
      e->param.operator.args[0] = NULL;
      e->param.operator.args[1] = NULL;
      if ((r = element_alloc_values(e, LABELCALL_RETVAL_MAXNUM)) < 0)
	return r;
      break;

    case ELEMENT_TYPE_NAME: /* variable or function name */
      if ((r = function_search(e->param.name.name, &function)) < 0) { /* variable */
	if ((r = variable_get(e->param.name.name, &variable)) < 0)
	  return r;
	e->type = ELEMENT_TYPE_VARIABLE;
	e->param.variable.variable = variable;
	if ((r = element_alloc_values(e, 1)) < 0)
	  return r;
      } else { /* function */
	e->type = ELEMENT_TYPE_OPERATOR;
	e->param.operator.op = operator_get(OPERATOR_TYPE_FUNCTION);
	e->param.operator.function = function;
	e->param.operator.label    = NULL;
	e->param.operator.args[0] = NULL;
	e->param.operator.args[1] = NULL;
	if ((r = element_alloc_values(e, function->retval_number)) < 0)
	  return r;
      }
      break;

    default:
      break;
    }
  }

  el = NULL;
  for (e = *elementp; e; el = e, e = e->next) {
    er = e->next;

    if ((e->type != ELEMENT_TYPE_OPERATOR) ||
	(e->param.operator.op->type != OPERATOR_TYPE_COMMA))
      continue;

    if (!el ||
	((el->type == ELEMENT_TYPE_OPERATOR) &&
	 (el->param.operator.op->type == OPERATOR_TYPE_COMMA))) {
      elp = el ? &(el->next) : elementp;
      if ((r = element_alloc(&el)) < 0)
	return r;
      el->type = ELEMENT_TYPE_OPERATOR;
      el->param.operator.op = operator_get(OPERATOR_TYPE_NOP);

      el->next = e;
      *elp = el;
    }

    if (!er ||
	((er->type == ELEMENT_TYPE_OPERATOR) &&
	 (er->param.operator.op->type == OPERATOR_TYPE_COMMA))) {
      el = e;
      if ((r = element_alloc(&e)) < 0)
	return r;
      e->type = ELEMENT_TYPE_OPERATOR;
      e->param.operator.op = operator_get(OPERATOR_TYPE_NOP);

      e->next = er;
      el->next = e;
    }
  }

  for (i = 0; i < OPERATOR_PRIORITY_NUM; i++) {
    priority = OPERATOR_PRIORITY_MIN + i;

    for (d = OPERATOR_DIRECTION_RL; d <= OPERATOR_DIRECTION_LR; d++) {

      e = *elementp;
      *elementp = NULL;
      for (; e; e = er) {
	er = e->next;
	e->next = *elementp;
	*elementp = e;
      }

      el = NULL;
      elp = NULL;
      for (e = *elementp; e; el = e, e = e->next) {
	er = e->next;

	if (e->type != ELEMENT_TYPE_OPERATOR)
	  goto next;
	if (e->param.operator.op->direction != d)
	  goto next;
	if (e->param.operator.op->priority != priority)
	  goto next;

	opetype = e->param.operator.op->type;

	switch (e->param.operator.op->subtype) {
	case OPERATOR_SUBTYPE_MON_BIN:
	  if (er && er->value) {
	    switch (opetype) {
	    case OPERATOR_TYPE_PLUS:
	      e->param.operator.op = operator_get(OPERATOR_TYPE_ADD); break;
	    case OPERATOR_TYPE_MINUS:
	      e->param.operator.op = operator_get(OPERATOR_TYPE_SUB); break;
	    case OPERATOR_TYPE_POINTER:
	      e->param.operator.op = operator_get(OPERATOR_TYPE_MUL); break;
	    case OPERATOR_TYPE_ADDRESS:
	      e->param.operator.op = operator_get(OPERATOR_TYPE_AND); break;
	    default:
	      return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
	    }
	    break;
	  }
	  /* Fall through */

	case OPERATOR_SUBTYPE_MONADIC:
	  if (!el)
	    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
	  *elp = e;
	  el->next = NULL;
	  e->param.operator.args[0] = el;
	  if ((r = element_alloc_values(e, e->param.operator.args[0]->value_num)) < 0)
	    return r;
	  continue;

	case OPERATOR_SUBTYPE_BINARY:
	  if (!el || !er)
	    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
	  *elp = e;
	  e->next = er->next;
	  el->next = NULL;
	  er->next = NULL;
	  switch (d) {
	  case OPERATOR_DIRECTION_RL:
	    e->param.operator.args[0] = er;
	    e->param.operator.args[1] = el;
	    break;
	  case OPERATOR_DIRECTION_LR:
	  default:
	    e->param.operator.args[0] = el;
	    e->param.operator.args[1] = er;
	    break;
	  }
	  if ((r = element_alloc_values(e, e->param.operator.args[0]->value_num)) < 0)
	    return r;
	  continue;

	case OPERATOR_SUBTYPE_NONE:
	  goto next;

	default:
	  return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
	}

	next:
	elp = elp ? &((*elp)->next) : elementp;
      }
    }
  }

  elp = elementp;
  for (el = *elementp; el && el->next; el = e) {
    e = el->next;

    if (((el->type == ELEMENT_TYPE_OPERATOR) &&
	 (el->param.operator.op->type == OPERATOR_TYPE_COMMA)) ||
	((e->type == ELEMENT_TYPE_OPERATOR) &&
	 (e->param.operator.op->type == OPERATOR_TYPE_COMMA))) {
      elp = &((*elp)->next);
      continue;
    }

    er = e;

    if ((r = element_alloc(&e)) < 0)
      return r;
    e->type = ELEMENT_TYPE_OPERATOR;
    e->param.operator.op = operator_get(OPERATOR_TYPE_CONT);

    *elp = e;
    e->next = er->next;
    el->next = NULL;
    er->next = NULL;
    e->param.operator.args[0] = el;
    e->param.operator.args[1] = er;
    if ((r = element_alloc_values(e, e->param.operator.args[1]->value_num)) < 0)
      return r;
  }

  for (e = *elementp; e && e->next; e = e->next) {
    er = e->next;
    if ((er->type == ELEMENT_TYPE_OPERATOR) &&
	(er->param.operator.op->type == OPERATOR_TYPE_COMMA)) {
      e->next = er->next;
      er->next = NULL;
      element_free(er);
    }
  }

  return 0;
}

static int proc_monadic_integer(operator_type_t opetype, value_t v, value_t v0)
{
  int r;
  integer_t x;

  if ((r = value_get_integer(v0, &x)) < 0)
    return r;

  switch (opetype) {
  case OPERATOR_TYPE_INC: x++; goto have_a_change;
  case OPERATOR_TYPE_DEC: x--; goto have_a_change;
    have_a_change:
    if ((r = value_set_integer(v0, x)) < 0)
      return r;
    break;

  case OPERATOR_TYPE_PLUS:          break;
  case OPERATOR_TYPE_MINUS: x = -x; break;
  case OPERATOR_TYPE_INV:   x = ~x; break;

  case OPERATOR_TYPE_NOT:
  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
  }

  if ((r = value_set_integer(v, x)) < 0)
    return r;

  return 0;
}

#ifdef NLL_FLOATING_POINT
static int proc_monadic_float(operator_type_t opetype, value_t v, value_t v0)
{
  int r;
  integer_t x;
  double f0;

  if ((r = value_get_float(v0, &f0)) < 0) {
    if ((r = value_get_integer(v0, &x)) < 0)
      return r;
    f0 = x;
  }

  switch (opetype) {
  case OPERATOR_TYPE_PLUS:            break;
  case OPERATOR_TYPE_MINUS: f0 = -f0; break;

  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
  }

  if ((r = value_set_float(v, f0)) < 0)
    return r;

  return 0;
}
#endif

static int proc_monadic_pointer(operator_type_t opetype, value_t v, value_t v0)
{
  int r, offset0;
  value_t p0;

  if ((r = value_get_pointer(v0, &p0, &offset0)) < 0)
    return r;

  switch (opetype) {
  case OPERATOR_TYPE_INC: offset0++; break;
  case OPERATOR_TYPE_DEC: offset0--; break;

  case OPERATOR_TYPE_NOT:
  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
  }

  if ((r = value_set_pointer(v0, p0, offset0)) < 0)
    return r;
  if ((r = value_set_pointer(v, p0, offset0)) < 0)
    return r;

  return 0;
}

static int proc_binary_integer(operator_type_t opetype, value_t v, value_t v0, value_t v1)
{
  int r;
  integer_t x, y;

  if ((r = value_get_integer(v0, &x)) < 0)
    return r;
  if ((r = value_get_integer(v1, &y)) < 0)
    return r;

  switch (opetype) {
  case OPERATOR_TYPE_DIV:
  case OPERATOR_TYPE_MOD:
    if (y == 0)
      return NLL_ERRCODE_FORMULA_DIV_ZERO;
    break;
  default:
    break;
  }

  switch (opetype) {
  case OPERATOR_TYPE_LSHIFT: x = x << y; break;
  case OPERATOR_TYPE_RSHIFT: x = x >> y; break;
  case OPERATOR_TYPE_LE:     x = x <= y; break;
  case OPERATOR_TYPE_GE:     x = x >= y; break;
  case OPERATOR_TYPE_EQ:     x = x == y; break;
  case OPERATOR_TYPE_NE:     x = x != y; break;
  case OPERATOR_TYPE_MUL:    x = x *  y; break;
  case OPERATOR_TYPE_DIV:    x = x /  y; break;
  case OPERATOR_TYPE_MOD:    x = x %  y; break;
  case OPERATOR_TYPE_ADD:    x = x +  y; break;
  case OPERATOR_TYPE_SUB:    x = x -  y; break;
  case OPERATOR_TYPE_LT:     x = x <  y; break;
  case OPERATOR_TYPE_GT:     x = x >  y; break;
  case OPERATOR_TYPE_XOR:    x = x ^  y; break;
  case OPERATOR_TYPE_AND:    x = x &  y; break;
  case OPERATOR_TYPE_OR:     x = x |  y; break;

  case OPERATOR_TYPE_LOGAND:
  case OPERATOR_TYPE_LOGOR:
  case OPERATOR_TYPE_SUBST:
  case OPERATOR_TYPE_REFER:
  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
  }

  if ((r = value_set_integer(v, x)) < 0)
    return r;

  return 0;
}

static int proc_binary_string(operator_type_t opetype, value_t v, value_t v0, value_t v1)
{
  int r, length0, length1;
  integer_t x;
  char *s0, *s1;
  char word[STRING_MAXLEN + 1];

  if ((r = value_get_string(v0, &s0, &length0)) < 0)
    return r;
  if ((r = value_get_string(v1, &s1, &length1)) < 0)
    return r;

  switch (opetype) {
  case OPERATOR_TYPE_ADD:
    if (length0 + length1 > STRING_MAXLEN)
      return NLL_ERRCODE_STRING_TOO_LONG;
    memcpy(word, s0, length0);
    memcpy(word + length0, s1, length1);
    word[length0 + length1] = '\0';
    if ((r = value_set_string(v, word, length0 + length1)) < 0)
      return r;
    return 0;

  case OPERATOR_TYPE_LT: x = (strcmp(s0, s1) <  0) ? 1 : 0; break;
  case OPERATOR_TYPE_GT: x = (strcmp(s0, s1) >  0) ? 1 : 0; break;
  case OPERATOR_TYPE_LE: x = (strcmp(s0, s1) <= 0) ? 1 : 0; break;
  case OPERATOR_TYPE_GE: x = (strcmp(s0, s1) >= 0) ? 1 : 0; break;
  case OPERATOR_TYPE_EQ: x = (strcmp(s0, s1) == 0) ? 1 : 0; break;
  case OPERATOR_TYPE_NE: x = (strcmp(s0, s1) != 0) ? 1 : 0; break;

  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
  }

  if ((r = value_set_integer(v, x)) < 0)
    return r;

  return 0;
}

#ifdef NLL_FLOATING_POINT
static int proc_binary_float(operator_type_t opetype, value_t v, value_t v0, value_t v1)
{
  int r;
  integer_t x, y;
  double f0, f1;

  if ((r = value_get_float(v0, &f0)) < 0) {
    if ((r = value_get_integer(v0, &x)) < 0)
      return r;
    f0 = x;
  }

  if ((r = value_get_float(v1, &f1)) < 0) {
    if ((r = value_get_integer(v1, &y)) < 0)
      return r;
    f1 = y;
  }

  switch (opetype) {
  case OPERATOR_TYPE_DIV:
  case OPERATOR_TYPE_MOD:
    if (f1 == 0)
      return NLL_ERRCODE_FORMULA_DIV_ZERO;
    break;
  default:
    break;
  }

  switch (opetype) {
  case OPERATOR_TYPE_LT: x = (f0 <  f1) ? 1 : 0; goto set_integer;
  case OPERATOR_TYPE_GT: x = (f0 >  f1) ? 1 : 0; goto set_integer;
  case OPERATOR_TYPE_LE: x = (f0 <= f1) ? 1 : 0; goto set_integer;
  case OPERATOR_TYPE_GE: x = (f0 >= f1) ? 1 : 0; goto set_integer;
  case OPERATOR_TYPE_EQ: x = (f0 == f1) ? 1 : 0; goto set_integer;
  case OPERATOR_TYPE_NE: x = (f0 != f1) ? 1 : 0; goto set_integer;
    set_integer:
    if ((r = value_set_integer(v, x)) < 0)
      return r;
    return 0;

  case OPERATOR_TYPE_MUL: f0 = f0 * f1; break;
  case OPERATOR_TYPE_DIV: f0 = f0 / f1; break;
  case OPERATOR_TYPE_ADD: f0 = f0 + f1; break;
  case OPERATOR_TYPE_SUB: f0 = f0 - f1; break;

  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
  }

  if ((r = value_set_float(v, f0)) < 0)
    return r;

  return 0;
}
#endif

static int proc_binary_pointer(operator_type_t opetype, value_t v, value_t v0, value_t v1)
{
  int r, offset0, offset1, is_int0 = 0, is_int1 = 0;
  integer_t x, integer0, integer1;
  value_t p0 = NULL, p1 = NULL;

  if ((r = value_get_pointer(v0, &p0, &offset0)) < 0) {
    if ((r = value_get_integer(v0, &integer0)) < 0)
      return r;
    is_int0 = 1;
  }

  if ((r = value_get_pointer(v1, &p1, &offset1)) < 0) {
    if ((r = value_get_integer(v1, &integer1)) < 0)
      return r;
    is_int1 = 1;
  }

  switch (opetype) {
  case OPERATOR_TYPE_ADD:
    if (p0 && is_int1) {
      offset1 = integer1;
    } else if (p1 && is_int0) {
      p0 = p1;
      offset0 = integer0;
    } else {
      return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
    }
    if ((r = value_set_pointer(v, p0, offset0 + offset1)) < 0)
      return r;
    return 0;

  case OPERATOR_TYPE_SUB:
    if (p0 && p1) {
      if ((r = value_set_integer(v, offset0 - offset1)) < 0)
	return r;
      return 0;
    } else if (p0 && is_int1) {
      offset1 = integer1;
    } else {
      return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
    }
    if ((r = value_set_pointer(v, p0, offset0 - offset1)) < 0)
      return r;
    return 0;

  default:
    break;
  }

  if (is_int0 || is_int1)
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;

  switch (opetype) {
  case OPERATOR_TYPE_LT: x = (p0 == p1) && (offset0 <  offset1); break;
  case OPERATOR_TYPE_GT: x = (p0 == p1) && (offset0 >  offset1); break;
  case OPERATOR_TYPE_LE: x = (p0 == p1) && (offset0 <= offset1); break;
  case OPERATOR_TYPE_GE: x = (p0 == p1) && (offset0 >= offset1); break;
  case OPERATOR_TYPE_EQ: x = (p0 == p1) && (offset0 == offset1); break;
  case OPERATOR_TYPE_NE: x = (p0 != p1) || (offset0 != offset1); break;

  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
  }

  if ((r = value_set_integer(v, x)) < 0)
    return r;

  return 0;
}

static int proc_monadic(element_t e)
{
  int r, offset;
  operator_type_t opetype;
  value_type_t t0;
  value_t v, v0, p;

  opetype = e->param.operator.op->type;

  v = e->value;
  v0 = e->param.operator.args[0]->value;

  for (; v && v0; v = v->next, v0 = v0->next) {
    switch (opetype) {
    case OPERATOR_TYPE_NOT:
      if ((r = value_is_nzero(v0)) < 0)
	return r;
      if ((r = value_set_integer(v, !r)) < 0)
	return r;
      continue;

    case OPERATOR_TYPE_POINTER:
      if ((r = value_get_pointer(v0, &p, &offset)) < 0)
	return r;
      if (!p)
	return NLL_ERRCODE_FORMULA_NULL_POINTER;
      if (offset < 0)
	return NLL_ERRCODE_FORMULA_INVALID_POINTER;
      for (; offset > 0; offset--) {
	if (!p->next)
	  return NLL_ERRCODE_FORMULA_INVALID_POINTER;
	p = p->next;
      }
      if ((r = value_set_value(v, p)) < 0)
	return r;
      continue;

    case OPERATOR_TYPE_ADDRESS:
      if ((r = value_set_pointer(v, v0, 0)) < 0)
	return r;
      continue;

    default:
      break;
    }

    t0 = value_get_type(v0);

#ifdef NLL_FLOATING_POINT
    if (t0 == VALUE_TYPE_FLOAT) {
      if ((r = proc_monadic_float(opetype, v, v0)) < 0)
	return r;
      continue;
    }
#endif

    if (t0 == VALUE_TYPE_POINTER) {
      if ((r = proc_monadic_pointer(opetype, v, v0)) < 0)
	return r;
      continue;
    }

    if ((r = proc_monadic_integer(opetype, v, v0)) < 0)
      return r;
  }

  return 0;
}

static int proc_binary(element_t e)
{
  int r, x;
  operator_type_t opetype;
  value_type_t t0, t1;
  value_t v, v0, v1;

  opetype = e->param.operator.op->type;

  v = e->value;
  v0 = e->param.operator.args[0]->value;
  v1 = e->param.operator.args[1]->value;

  for (; v && v0 && v1; v = v->next, v0 = v0->next, v1 = v1->next) {
    switch (opetype) {
    case OPERATOR_TYPE_CONT:
      if ((r = value_is_nzero(v0)) < 0)
	return r;
      if (!r) {
	if ((r = value_set_value(v, v0)) < 0)
	  return r;
      } else {
	if ((r = value_set_value(v, v1)) < 0)
	  return r;
      }
      continue;

    case OPERATOR_TYPE_LOGAND:
      if ((r = value_is_nzero(v0)) < 0)
	return r;
      x = r;
#ifndef NLL_LOGICAL_OPERATOR_ORDER
      r = 1;
#endif
      if (r) {
	if ((r = value_is_nzero(v1)) < 0)
	  return r;
	x = x && r;
      }
      if ((r = value_set_integer(v, x)) < 0)
	return r;
      continue;

    case OPERATOR_TYPE_LOGOR:
      if ((r = value_is_nzero(v0)) < 0)
	return r;
      x = r;
#ifndef NLL_LOGICAL_OPERATOR_ORDER
      r = 0;
#endif
      if (!r) {
	if ((r = value_is_nzero(v1)) < 0)
	  return r;
	x = x || r;
      }
      if ((r = value_set_integer(v, x)) < 0)
	return r;
      continue;

    case OPERATOR_TYPE_REFER:
      if ((r = value_set_value_value(v0, v1)) < 0)
	return r;
      if ((r = value_set_value(v, v0)) < 0)
	return r;
      continue;

    default:
      break;
    }

    t0 = value_get_type(v0);
    t1 = value_get_type(v1);

    if ((t0 == VALUE_TYPE_STRING) || (t1 == VALUE_TYPE_STRING)) {
      if ((r = proc_binary_string(opetype, v, v0, v1)) < 0)
	return r;
      continue;
    }

#ifdef NLL_FLOATING_POINT
    if ((t0 == VALUE_TYPE_FLOAT) || (t1 == VALUE_TYPE_FLOAT)) {
      if ((r = proc_binary_float(opetype, v, v0, v1)) < 0)
	return r;
      continue;
    }
#endif

    if ((t0 == VALUE_TYPE_POINTER) || (t1 == VALUE_TYPE_POINTER)) {
      if ((r = proc_binary_pointer(opetype, v, v0, v1)) < 0)
	return r;
      continue;
    }

    if ((r = proc_binary_integer(opetype, v, v0, v1)) < 0)
      return r;
  }

  return 0;
}

static int label_proc(label_t label, value_t value, value_t args)
{
  int r, errcode = 0;
  struct spot s;
  value_t retval;

  if (!label || !label->spot.line)
    return NLL_ERRCODE_LABEL_NOT_FOUND;

  s.line    = label->spot.line;
  s.command = label->spot.command;

  if ((r = formula_proc(s.command->args[0].arg.label.args, NULL)) < 0)
    return r;
  if (s.command->args[0].arg.label.args) {
    if ((r = value_copy_values(s.command->args[0].arg.label.args->value, args)) < 0)
      return r;
  }

  if ((r = position_push(POSITION_TYPE_LABEL, NULL, NULL)) < 0)
    return r;

  if ((r = nll_run(&s)) < 0)
    return r;

  if (r > 0)
    errcode = NLL_ERRCODE_NLL_INVALID_COMMAND;

  if ((r = position_pop(POSITION_TYPE_LABEL, NULL, &retval)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_copy_values(value, retval)) < 0)
    return r;

  return errcode;
}

static int proc(element_t element, value_t *valuep);

static int proc_element(element_t element)
{
  int r, n;
  integer_t integer;
  operator_type_t opetype;
  array_t array;
  element_t e;
  value_t value;

  if (element->type == ELEMENT_TYPE_VARIABLE) {
    if ((r = value_set_value(element->value, element->param.variable.variable->value)) < 0)
      return r;
  }

  if (element->type == ELEMENT_TYPE_ELEMENTS) {
    if ((r = proc(element->param.elements.head, NULL)) < 0)
      return r;
    if ((r = element_set_values(element, element->param.elements.head)) < 0)
      return r;
  }

  if (element->type != ELEMENT_TYPE_OPERATOR)
    return 0;

  opetype = element->param.operator.op->type;

  switch (element->param.operator.op->subtype) {
  case OPERATOR_SUBTYPE_MON_BIN:
  case OPERATOR_SUBTYPE_MONADIC:
    if (!element->param.operator.args[0])
      return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

    if ((r = proc(element->param.operator.args[0], NULL)) < 0)
      return r;

    if (opetype == OPERATOR_TYPE_FUNCTION) {
      if (element->param.operator.args[0]->type != ELEMENT_TYPE_ELEMENTS)
	return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
      if ((r = function_proc(element->param.operator.function, element->value,
			     element->param.operator.args[0]->value)) < 0)
	return r;
      break;
    }

    if (opetype == OPERATOR_TYPE_LABELCALL) {
      if (element->param.operator.args[0]->type != ELEMENT_TYPE_ELEMENTS)
	return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
      if ((r = label_proc(element->param.operator.label, element->value,
			  element->param.operator.args[0]->value)) < 0)
	return r;
      break;
    }

    if ((r = proc_monadic(element)) < 0)
      return r;

    break;

  case OPERATOR_SUBTYPE_BINARY:
    if (!element->param.operator.args[0] ||
	!element->param.operator.args[1])
      return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

    if ((r = proc(element->param.operator.args[0], NULL)) < 0)
      return r;

    switch (opetype) {
    case OPERATOR_TYPE_CONT:
#ifdef NLL_LOGICAL_OPERATOR_ORDER
    case OPERATOR_TYPE_LOGAND:
#endif
      value = element->param.operator.args[0]->value;
      e = element->param.operator.args[1];
      if (e->type == ELEMENT_TYPE_ELEMENTS) {
	if ((r = element_set_values(e, e->param.elements.head)) < 0)
	  return r;
	e = e->param.elements.head;
      }
      for (; value && e; value = value->next, e = e->next) {
	if ((r = value_is_nzero(value)) < 0)
	  return r;
	if (r) {
	  if ((r = proc_element(e)) < 0)
	    return r;
	}
      }
      break;

#ifdef NLL_LOGICAL_OPERATOR_ORDER
    case OPERATOR_TYPE_LOGOR:
      value = element->param.operator.args[0]->value;
      e = element->param.operator.args[1];
      if (e->type == ELEMENT_TYPE_ELEMENTS) {
	if ((r = element_set_values(e, e->param.elements.head)) < 0)
	  return r;
	e = e->param.elements.head;
      }
      for (; value && e; value = value->next, e = e->next) {
	if ((r = value_is_nzero(value)) < 0)
	  return r;
	if (!r) {
	  if ((r = proc_element(e)) < 0)
	    return r;
	}
      }
      break;
#endif

    default:
      if ((r = proc(element->param.operator.args[1], NULL)) < 0)
	return r;
      break;
    }

    if (opetype == OPERATOR_TYPE_SUBST) {
      /*
       * (x, y) = (y, x):
       * To avoid writing x before reading x,
       * copy source to result and result to destination.
       */
      if ((r = value_copy_values(element->value, element->param.operator.args[1]->value)) < 0)
	return r;
      if ((r = value_copy_values(element->param.operator.args[0]->value, element->value)) < 0)
	return r;
      break;
    }

    if (opetype == OPERATOR_TYPE_ARRAY) {
      if ((r = value_get_integer(element->param.operator.args[1]->value, &integer)) < 0)
	return r;
      if ((r = value_get_array(element->param.operator.args[0]->value, &n, &array)) < 0)
	return r;
      if (integer >= n)
	return NLL_ERRCODE_FORMULA_ARRAY_OVER;
      if ((r = value_set_value(element->value, array[integer].value)) < 0)
	return r;
      break;
    }

    if ((r = proc_binary(element)) < 0)
      return r;

    break;

  case OPERATOR_SUBTYPE_NONE:
    break;

  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR;
  }

  return 0;
}

static int proc(element_t element, value_t *valuep)
{
  int r;
  element_t e;
  value_t evalue = NULL;

  for (e = element; e; e = e->next) {
    if ((r = proc_element(e)) < 0)
      return r;
    evalue = e->value;
  }

  if (valuep)
    *valuep = evalue;

  return 0;
}

int formula_free(element_t *elementp)
{
  element_t e;

  while (*elementp) {
    e = *elementp;
    *elementp = (*elementp)->next;
    e->next = NULL;

    switch (e->type) {
    case ELEMENT_TYPE_VARIABLE:
      if (e->param.variable.variable)
	variable_del(e->param.variable.variable);
      break;

    case ELEMENT_TYPE_ELEMENTS:
      if (e->param.elements.head)
	formula_free(&e->param.elements.head);
      break;

    case ELEMENT_TYPE_OPERATOR:
      if (e->param.operator.label)
	label_del(e->param.operator.label);
      if (e->param.operator.args[0])
	formula_free(&e->param.operator.args[0]);
      if (e->param.operator.args[1])
	formula_free(&e->param.operator.args[1]);
      break;

    default:
      break;
    }

    element_free(e);
  }

  return 0;
}

int formula_clean(element_t element)
{
  int r;

  for (; element; element = element->next) {
    if ((r = element_clean(element)) < 0)
      return r;
    switch (element->type) {
    case ELEMENT_TYPE_VARIABLE:
      if (element->param.variable.variable) {
	if ((r = variable_clean(element->param.variable.variable)) < 0)
	  return r;
      }
      break;

    case ELEMENT_TYPE_ELEMENTS:
      if (element->param.elements.head) {
	if ((r = formula_clean(element->param.elements.head)) < 0)
	  return r;
      }
      break;

    case ELEMENT_TYPE_OPERATOR:
      if (element->param.operator.label) {
	if ((r = label_clean(element->param.operator.label)) < 0)
	  return r;
      }
      if (element->param.operator.args[0]) {
	if ((r = formula_clean(element->param.operator.args[0])) < 0)
	  return r;
      }
      if (element->param.operator.args[1]) {
	if ((r = formula_clean(element->param.operator.args[1])) < 0)
	  return r;
      }
      break;

    default:
      break;
    }
  }

  return 0;
}

int formula_parse(char *formula, char **endp, element_t *elementp, const char *terminator)
{
  element_t element = NULL;
  int r;

  if (!terminator)
    terminator = ",;";

  if ((r = lex(&element, formula, endp, terminator)) < 0) {
    formula_free(&element);
    return r;
  }

  if ((r = parse(&element)) < 0) {
    formula_free(&element);
    return r;
  }

  if (elementp)
    *elementp = element;

  return 0;
}

int formula_proc(element_t element, value_t *valuep)
{
  int r;
  value_t value;

  if ((r = proc(element, &value)) < 0)
    return r;

  if (valuep)
    *valuep = value;

  return 0;
}

int formula_dump(element_t element)
{
  const char *name;
  element_t left = NULL;

  for (; element; left = element, element = element->next) {
    if (left)
      printf(",");
    switch (element->type) {
    case ELEMENT_TYPE_INTEGER:
    case ELEMENT_TYPE_STRING:
#ifdef NLL_FLOATING_POINT
    case ELEMENT_TYPE_FLOAT:
#endif
      value_dump(element->value);
      break;

    case ELEMENT_TYPE_VARIABLE:
      printf("%s", element->param.variable.variable->name);
      break;

    case ELEMENT_TYPE_ELEMENTS:
      printf("(");
      formula_dump(element->param.elements.head);
      printf(")");
      break;

    case ELEMENT_TYPE_OPERATOR:
      name = "";
      if (element->param.operator.op->word) {
	name = element->param.operator.op->word;
      } else if (element->param.operator.op->type == OPERATOR_TYPE_CONT) {
	name = " ";
      } else if (element->param.operator.op->type == OPERATOR_TYPE_NOP) {
	name = "NOP";
      } else if (element->param.operator.op->type == OPERATOR_TYPE_FUNCTION) {
	name = function_name(element->param.operator.function);
      } else if (element->param.operator.op->type == OPERATOR_TYPE_LABELCALL) {
	name = element->param.operator.label->name;
      }

      switch (element->param.operator.op->subtype) {
      case OPERATOR_SUBTYPE_MONADIC:
      case OPERATOR_SUBTYPE_MON_BIN:
	printf("%s", name);
	if (element->param.operator.args[0])
	  formula_dump(element->param.operator.args[0]);
	break;

      case OPERATOR_SUBTYPE_BINARY:
	if (element->param.operator.args[0])
	  formula_dump(element->param.operator.args[0]);
	printf("%s", name);
	if (element->param.operator.args[1])
	  formula_dump(element->param.operator.args[1]);
	if (element->param.operator.op->type == OPERATOR_TYPE_ARRAY)
	  printf("]");
	break;

      case OPERATOR_SUBTYPE_NONE:
      default:
	printf("%s", name);
	break;
      }
      break;

    case ELEMENT_TYPE_NONE:
    default:
      printf("?");
      break;
    }
  }

  return 0;
}
