#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "lib.h"
#include "string.h"
#include "array.h"
#include "area.h"
#include "framebuf.h"
#include "value.h"

static struct value values[VALUE_MAXNUM];

static value_t pool = NULL;

int value_check(void)
{
  value_t value;
  int n = 0;

  for (value = pool; value && (n < VALUE_MAXNUM); value = value->next)
    n++;

  if (value || (n != VALUE_MAXNUM))
    return NLL_ERRCODE_VALUE_NOT_EMPTY;

  return 0;
}

int value_alloc(value_t *valuep)
{
  value_t value;

  if (!pool)
    return NLL_ERRCODE_VALUE_BUFFER_OVER;

  value = pool;
  pool = pool->next;

  memset(value, 0, sizeof(*value));

  value->refcount++;

  *valuep = value;

  return 0;
}

int value_free(value_t value)
{
  value_t next;
  int r;

  if (value->refcount > 1) {
    value->refcount--;
    return 0;
  }

  for (; value; value = next) {
    next = value->next;
    value->next = NULL;

    if (--(value->refcount) == 0) {
      if (value_is_const(value)) {
	if ((r = value_reset_const(value)) < 0)
	  return r;
      }
      if ((r = value_clear(value)) < 0)
	return r;

      value->next = pool;
      pool = value;
    }
  }

  return 0;
}

int value_init(void)
{
  value_t value;
  int i;

  pool = NULL;
  memset(values, 0, sizeof(values));

  for (i = 0; i < VALUE_MAXNUM; i++) {
    value = &values[i];
    value->next = pool;
    pool = value;
  }

  return 0;
}

int value_clear(value_t value)
{
  int r;

  if (value_is_const(value))
    return 0;

  switch (value->type) {
  case VALUE_TYPE_ARRAY:
    if ((r = array_free(value->val.array.array, value->val.array.number)) < 0)
      return r;
    break;

  case VALUE_TYPE_STRING:
    if ((r = string_del(value->val.string)) < 0)
      return r;
    break;

  case VALUE_TYPE_VALUE:
    if ((r = value_free(value->val.value)) < 0)
      return r;
    break;

  case VALUE_TYPE_AREA:
    if ((r = area_free(value->val.area)) < 0)
      return r;
    break;

  case VALUE_TYPE_POINTER:
    if (value->val.pointer.point) {
      if ((r = value_free(value->val.pointer.point)) < 0)
	return r;
    }
    break;

  default:
    break;
  }

  value->type = VALUE_TYPE_UNDEFINED;

  return 0;
}

int value_is_const(value_t value)
{
  return (value->flags & VALUE_FLAG_CONST) ? 1 : 0;
}

int value_set_const(value_t value)
{
  value->flags |= VALUE_FLAG_CONST;
  return 0;
}

int value_reset_const(value_t value)
{
  value->flags &= ~VALUE_FLAG_CONST;
  return 0;
}

int value_is_nzero(value_t value)
{
  int r, length;
  value_type_t type;
  integer_t integer;
  char *string;
#ifdef NLL_FLOATING_POINT
  double f;
#endif
  area_t area;
  value_t point;

  type = value_get_type(value);

  switch (type) {
  case VALUE_TYPE_INTEGER:
    if ((r = value_get_integer(value, &integer)) < 0)
      return r;
    return integer ? 1 : 0;

  case VALUE_TYPE_STRING:
    if ((r = value_get_string(value, &string, &length)) < 0)
      return r;
    return length ? 1 : 0;

#ifdef NLL_FLOATING_POINT
  case VALUE_TYPE_FLOAT:
    if ((r = value_get_float(value, &f)) < 0)
      return r;
    return (f != 0.0) ? 1 : 0;
#endif

  case VALUE_TYPE_AREA:
    if ((r = value_get_area(value, &area)) < 0)
      return r;
    return area ? 1 : 0;

  case VALUE_TYPE_POINTER:
    if ((r = value_get_pointer(value, &point, NULL)) < 0)
      return r;
    return point ? 1 : 0;

  default:
    break;
  }

  return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
}

int value_copy_values(value_t dvalue, value_t svalue)
{
  int r;
  value_t value;

  for (; dvalue && svalue; dvalue = dvalue->next, svalue = svalue->next) {
    for (value = svalue; value->type == VALUE_TYPE_VALUE; value = value->val.value)
      ;
    switch (value->type) {
    case VALUE_TYPE_INTEGER:
      if ((r = value_set_integer(dvalue, value->val.integer)) < 0)
	return r;
      break;
    case VALUE_TYPE_STRING:
      if ((r = value_set_string(dvalue, value->val.string->s, value->val.string->length)) < 0)
	return r;
      break;
#ifdef NLL_FLOATING_POINT
    case VALUE_TYPE_FLOAT:
      if ((r = value_set_float(dvalue, value->val.floating)) < 0)
	return r;
      break;
#endif
    case VALUE_TYPE_AREA:
      if ((r = value_set_area(dvalue, value->val.area)) < 0)
	return r;
      break;
    case VALUE_TYPE_POINTER:
      if ((r = value_set_pointer(dvalue, value->val.pointer.point, value->val.pointer.offset)) < 0)
	return r;
      break;
    case VALUE_TYPE_UNDEFINED:
      break;
    case VALUE_TYPE_ARRAY:
    default:
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
    }
  }

  return 0;
}

value_type_t value_get_type(value_t value)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  return value->type;
}

int value_is_same(value_t value0, value_t value1)
{
  while (value0->type == VALUE_TYPE_VALUE)
    value0 = value0->val.value;
  while (value1->type == VALUE_TYPE_VALUE)
    value1 = value1->val.value;
  return (value0 == value1) ? 1 : 0;
}

int value_get_value(value_t value, value_t *valuep)
{
  if (value->type != VALUE_TYPE_VALUE)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *valuep = value->val.value;

  return 0;
}

int value_set_value(value_t value, value_t _value)
{
  int r;

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_VALUE))
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((value->type != VALUE_TYPE_UNDEFINED) && value_is_const(value))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  _value->refcount++;

  if (value->type == VALUE_TYPE_VALUE) {
    if ((r = value_free(value->val.value)) < 0)
      return r;
  }

  value->type = VALUE_TYPE_VALUE;
  value->val.value = _value;

  return 0;
}

int value_set_value_value(value_t value, value_t _value)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;
  return value_set_value(value, _value);
}

int value_clear_value(value_t value)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;
  return value_clear(value);
}

int value_set_values(value_t value, value_t _value)
{
  int r;

  for (; value && _value; value = value->next, _value = _value->next) {
    if ((r = value_set_value(value, _value)) < 0)
      return r;
  }

  return 0;
}

int value_get_array(value_t value, int *numberp, array_t *arrayp)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if (value->type != VALUE_TYPE_ARRAY)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *numberp = value->val.array.number;
  *arrayp = value->val.array.array;

  return 0;
}

int value_set_array(value_t value, int number, array_t array)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_ARRAY))
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((value->type != VALUE_TYPE_UNDEFINED) && value_is_const(value))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  value->type = VALUE_TYPE_ARRAY;
  value->val.array.number = number;
  value->val.array.array = array;

  return 0;
}

int value_get_integer(value_t value, integer_t *integerp)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if (value->type != VALUE_TYPE_INTEGER)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *integerp = value->val.integer;

  return 0;
}

int value_set_integer(value_t value, integer_t integer)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_INTEGER))
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((value->type != VALUE_TYPE_UNDEFINED) && value_is_const(value))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  value->type = VALUE_TYPE_INTEGER;
  value->val.integer = integer;

  return 0;
}

int value_get_string(value_t value, char **stringp, int *lengthp)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if (value->type != VALUE_TYPE_STRING)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *stringp = value->val.string->s;
  if (lengthp)
    *lengthp = value->val.string->length;

  return 0;
}

int value_set_string(value_t value, char *string, int length)
{
  int r;

  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_STRING))
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((value->type != VALUE_TYPE_UNDEFINED) && value_is_const(value))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  if (length > STRING_MAXLEN)
    return NLL_ERRCODE_STRING_TOO_LONG;

  if (value->type == VALUE_TYPE_UNDEFINED) {
    if ((r = string_alloc(&value->val.string)) < 0)
      return r;
  }

  value->type = VALUE_TYPE_STRING;
  if ((r = string_set(value->val.string, string, length)) < 0)
    return 0;

  return 0;
}

#ifdef NLL_FLOATING_POINT
int value_get_float(value_t value, double *floatingp)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if (value->type != VALUE_TYPE_FLOAT)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *floatingp = value->val.floating;

  return 0;
}

int value_set_float(value_t value, double floating)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_FLOAT))
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((value->type != VALUE_TYPE_UNDEFINED) && value_is_const(value))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  value->type = VALUE_TYPE_FLOAT;
  value->val.floating = floating;

  return 0;
}
#endif

int value_get_area(value_t value, area_t *areap)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if (value->type != VALUE_TYPE_AREA)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *areap = value->val.area;

  return 0;
}

int value_set_area(value_t value, area_t area)
{
  int r;

  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_AREA))
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((value->type != VALUE_TYPE_UNDEFINED) && value_is_const(value))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  if (area)
    area->refcount++;

  if (value->type == VALUE_TYPE_AREA) {
    if ((r = area_free(value->val.area)) < 0)
      return r;
  }

  value->type = VALUE_TYPE_AREA;
  value->val.area = area;

  return 0;
}

int value_get_pointer(value_t value, value_t *pointp, int *offsetp)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if (value->type != VALUE_TYPE_POINTER)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *pointp = value->val.pointer.point;
  if (offsetp)
    *offsetp = value->val.pointer.offset;

  return 0;
}

int value_set_pointer(value_t value, value_t point, int offset)
{
  int r;

  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_POINTER))
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((value->type != VALUE_TYPE_UNDEFINED) && value_is_const(value))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  if (point)
    point->refcount++;

  if (value->type == VALUE_TYPE_POINTER) {
    if (value->val.pointer.point) {
      if ((r = value_free(value->val.pointer.point)) < 0)
	return r;
    }
  }

  value->type = VALUE_TYPE_POINTER;
  value->val.pointer.point = point;
  value->val.pointer.offset = point ? offset : 0;

  return 0;
}

int value_print(value_t value)
{
  value_t v;
  char buf[32];

  for (v = value; v; v = v->next) {
    for (value = v; value->type == VALUE_TYPE_VALUE; value = value->val.value)
      ;
    switch (value->type) {
    case VALUE_TYPE_INTEGER:
      sprintf(buf, "%ld", value->val.integer);
      framebuf_puts(buf);
      break;
    case VALUE_TYPE_STRING:
      framebuf_puts(value->val.string->s);
      break;
#ifdef NLL_FLOATING_POINT
    case VALUE_TYPE_FLOAT:
      sprintf(buf, "%f", value->val.floating);
      framebuf_puts(buf);
      break;
#endif
    case VALUE_TYPE_ARRAY:
    case VALUE_TYPE_AREA:
    case VALUE_TYPE_POINTER:
    case VALUE_TYPE_UNDEFINED:
    default:
      break;
    }
  }

  return 0;
}

int value_dump(value_t value)
{
  value_t v = NULL;
  int i;

  for (v = value; value; value = value->next) {
    if (v != value)
      printf(",");
    if (value_is_const(value)) {
      switch (value->type) {
      case VALUE_TYPE_INTEGER:
      case VALUE_TYPE_STRING:
#ifdef NLL_FLOATING_POINT
      case VALUE_TYPE_FLOAT:
#endif
	break;
      default:
	printf("(C)");
      }
    }
    switch (value->type) {
    case VALUE_TYPE_VALUE:
      printf("(");
      value_dump(value->val.value);
      printf(")");
      break;
    case VALUE_TYPE_ARRAY:
      printf("ARRAY[%d](", value->val.array.number);
      for (i = 0; i < value->val.array.number; i++) {
	printf("[%d]", i);
	value_dump(value->val.array.array[i].value);
      }
      printf(")");
      break;
    case VALUE_TYPE_INTEGER:
      printf("%ld", value->val.integer);
      break;
    case VALUE_TYPE_STRING:
      string_print(value->val.string);
      break;
#ifdef NLL_FLOATING_POINT
    case VALUE_TYPE_FLOAT:
      printf("%f(FLOAT)", value->val.floating);
      break;
#endif
    case VALUE_TYPE_AREA:
      printf("AREA(%d)", value->val.area->size);
      area_print(value->val.area);
      break;
    case VALUE_TYPE_POINTER:
      printf("&");
      if (value->val.pointer.point) {
	value_dump(value->val.pointer.point);
	printf("[%d]", value->val.pointer.offset);
      } else {
	printf("NULL");
      }
      break;
    case VALUE_TYPE_UNDEFINED:
    default:
      printf("UNDEF");
      break;
    }
  }

  fflush(stdout);

  return 0;
}
